/*
  ESP_IRIP_Master.ino

  This sketch for receive infrared remote signals
   and transmits them over ESP-NOW to a slave device.

*/

#include "ESP32_NOW.h"
#include "WiFi.h"
#include <IRremoteESP8266.h>
#include <IRrecv.h>
#include <IRac.h>
#include <IRutils.h>

#define ESPNOW_WIFI_CHANNEL 6 
#define ESPNOW_MAX_PAYLOAD 250  
#define END_MARKER 0xFFFF

uint16_t end_marker[1] = {END_MARKER};

const uint16_t kRecvPin = 16; // Pin number connected to the IR receiver
const uint16_t kCaptureBufferSize = 1024; // Buffer size for capturing IR data
const uint8_t kTimeout = 12; // Timeout value for IR decoding
const uint16_t kMinUnknownSize = 15; // Minimum size for an unknown IR protocol
const uint8_t kTolerancePercentage = kTolerance;  // Tolerance percentage for IR decoding

IRrecv irrecv(kRecvPin, kCaptureBufferSize, kTimeout, true);
decode_results results;

class ESP_NOW_Broadcast_Peer : public ESP_NOW_Peer {
public:
  ESP_NOW_Broadcast_Peer(uint8_t channel, wifi_interface_t iface, const uint8_t *lmk) 
    : ESP_NOW_Peer(ESP_NOW.BROADCAST_ADDR, channel, iface, lmk) {}

  ~ESP_NOW_Broadcast_Peer() {
    remove();
  }

  bool begin() {
    if (!ESP_NOW.begin() || !add()) {
      log_e("Failed to initialize ESP-NOW or register the broadcast peer");
      return false;
    }
    return true;
  }

  bool send_message(const uint8_t *data, size_t len) {
    if (!send(data, len)) {
      log_e("Failed to broadcast message");
      return false;
    }
    return true;
  }
};

ESP_NOW_Broadcast_Peer broadcast_peer(ESPNOW_WIFI_CHANNEL, WIFI_IF_STA, NULL);

void sendIRdata(uint16_t *data, int len) {
  Serial.printf("[sendIRdata] Sending %d bytes of IR data:\n", len * sizeof(uint16_t));
  uint8_t buffer[ESPNOW_MAX_PAYLOAD];

  for (int offset = 0; offset < len * 2; offset += ESPNOW_MAX_PAYLOAD) {
    int bytes_to_send = min(len * 2 - offset, ESPNOW_MAX_PAYLOAD);

    memcpy(buffer, (uint8_t*)data + offset, bytes_to_send);

    if (!broadcast_peer.send_message(buffer, bytes_to_send)) {
      Serial.println("[sendIRdata] Failed to broadcast message");
    }
  }
}

void setup() {
  Serial.begin(115200);
  delay(1000);
  WiFi.mode(WIFI_STA);
  WiFi.setChannel(ESPNOW_WIFI_CHANNEL);
  while (!WiFi.STA.started()) {
    delay(100);
  }

  Serial.println("ESP-NOW - Broadcast Master");
  Serial.printf("  Channel: %d\n", ESPNOW_WIFI_CHANNEL);

  if (!broadcast_peer.begin()) {
    Serial.println("Failed to initialize broadcast peer");
    Serial.println("Rebooting in 5 seconds...");
    delay(5000);
    ESP.restart();
  }

  // Perform a low level sanity checks that the compiler performs bit field
  // packing as we expect and Endianness is as we expect.
  assert(irutils::lowLevelSanityCheck() == 0);

  irrecv.setTolerance(kTolerancePercentage);  // Override the default tolerance.
  irrecv.enableIRIn();  // Start the receiver

  sendIRdata(end_marker, 1);

  Serial.println("Setup complete.");
}

void handleIRData(const decode_results& results) {
  //Serial.print(resultToHumanReadableBasic(&results));
  //Serial.print(resultToSourceCode(&results));

  uint16_t* rawdata = resultToRawArray(&results);
  if (!rawdata) {
    Serial.println("[handleIRData] Failed to allocate memory for raw data.");
    return;
  }

  int rawlen = getCorrectedRawLength(&results);
  Serial.printf("[handleIRData] Raw length: %d\n", rawlen);

  sendIRdata(rawdata, rawlen);

  bool needsEndMarker = (rawlen * 2 % ESPNOW_MAX_PAYLOAD == 0);
  if (needsEndMarker) {
    Serial.println("[handleIRData] END_MARKER needed.");
    sendIRdata(end_marker, 1);
  }

  free(rawdata);
  Serial.println("");
}

void loop() {
  if (irrecv.decode(&results)) {
    if (results.overflow) {
        Serial.println("[loop] Overflow detected!");
    } else {
        handleIRData(results);
    }
    delay(20);
    irrecv.resume();
  }
}
